/*
 * File Name: xauthn.c
 *
 * Licensed Materials - Property of IBM
 * 5748-XX8
 * (c) Copyright International Business Machines Corp. 2002
 * All Rights Reserved.
 * US Government Users Restricted Rights - Use, duplicaion or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 */
/*
 * FILE: xauthn.c
 *
 * AM Plug-in cross domain authentication (CDAS) demo.  This file implements
 *
 *     xauthn_initialize()
 *     xauthn_shutdown()
 *     xauthn_authenticate()
 *     xauthn_change_password()
 *
 * Functions used by the AM Plug-in for Edge Server to authenticate
 * users based on the specified authentication mechanism.
 *
 * To use the CDAS demo shared object the following configuration 
 * directives must be modified in the osdef.conf file.
 *
 * [Global]
 * cdas_loaded = yes
 * cdas_init_parameter = demouser <userid>  demopassword <password>  democertdn <CertDN>  validpddn <validpdDN>
 * 
 * [Local]
 * #For the demo basic_auth  and cert are valid values for cdas_auth_parameter.
 * cdas_enabled = yes
 * cdas_auth_parameter = basic_auth
 *
 *
 * In the settings above the demo module is being configured with 
 * a userid, password, certificate distinguished  name, and a
 * valid AM LDAP distinguished name.  The arguments specified by the 
 * cdas_init_parameter will be passed to the demo module 
 * when the  *xauthn_initialize() function is called.
 */


#include "xauthn.h"

/*In a NON-demo environment you should protect global 
 variables to make sure this is thread safe.*/
demo_data Global_demo_info;


/*
 * FUNCTION NAME
 * 	xauthn_initialize
 *
 * DESCRIPTION
 * 	init the authentication service
 *
 * ARGUMENTS
 * 	[in] argc The count of arguments to the service.
 * 	[in] argv The array of argument strings.
 *
 * RETURN VALUE
 * 	XAUTHN_S_COMPLETE on success, error code on failure
 */
xauthn_status_t
xauthn_initialize(
    int        argc,     /* in */
    const char **argv    /* in */
)
{
    int found_du=0, found_dup=0, found_cdn=0,found_vpddn=0;
    int status=XAUTHN_S_COMPLETE;
    int argcount=0;
    int type=0;

    memset(Global_demo_info.demo_user,0,BIGDATASIZE);
    memset(Global_demo_info.demo_user_password,0,BIGDATASIZE);
    memset(Global_demo_info.demo_cert_dn,0,BIGDATASIZE);
    memset(Global_demo_info.demo_validpddn,0,BIGDATASIZE);


    /*Go through the arguments to make sure all the demo init data*/
    /*has been passed in. If not then return a negative error code*/
    for (argcount=0; argcount<argc; argcount++)
    {
       type=init_arg_type((char *)argv[argcount]);

       switch(type)
        {
		case DEMOUSER:
                          status=getOptionValue(&argcount,argc,(char **)argv,Global_demo_info.demo_user);
                          if (status == 0)
                          {
                             found_du = 1;
                          } 
                          break;
		case DEMOUSERPASS:
                          status=getOptionValue(&argcount,argc,(char **)argv,Global_demo_info.demo_user_password);
                          if (status == 0)
                          {
                             found_dup =1;
                          } 
                          break;
		case DEMOCERTDN:
                          status=getOptionValue(&argcount,argc,(char **)argv,Global_demo_info.demo_cert_dn);
                          if (status == 0)
                          {
                             found_cdn=1;
                          } 
                          break; 
		case VALIDPDDN:
                          status=getOptionValue(&argcount,argc,(char **)argv,Global_demo_info.demo_validpddn);
                          if (status == 0)
                          {
                             found_vpddn=1;
                          } 
                          break; 

        }

    }

    /*If all the demo data wasn't passed in then return a
       bad error code*/
    if (found_du==0 || found_dup==0 || found_cdn==0 || found_vpddn==0)
     {
        status=XAUTHN_S_FAILURE;
     }
    else{
        status=XAUTHN_S_COMPLETE;
     }

    return status;
}


/*
 * FUNCTION NAME
 * 	xauthn_shutdown
 *
 * DESCRIPTION
 * 	Shutdown the authentication service.
 * 	The initialization parameters are passed in
 * 	again.
 *
 * ARGUMENTS
 * 	[in] argc     The count of arguments to the service.
 * 	[in] argv     The array of argument strings.
 *
 * RETURN VALUE
 * 	XAUTHN_S_COMPLETE
 */
xauthn_status_t
xauthn_shutdown(
    int        argc,     /* in */
    const char **argv    /* in */
)
{
    return XAUTHN_S_COMPLETE;
}



/*
 * FUNCTION NAME
 *      xauthn_authenticate
 *
 * DESCRIPTION
 *      Examine the received user authentication information, and generate a 
 *      client identity. The received information will vary depending on the
 *      specified authentication mechanism. 
 *
 * ARGUMENTS - IN
 *      authInfo      List of names and set of values containing the user
 *                    authentication data. The pdxauthn.h contains all the
 *                    possible names that could be in this list. The actual
 *                    list of names received will depend on the specified    
 *                    authentication mechanism. 
 *
 * ARGUMENTS - OUT
 *      identity      Pointer to the resulted client identity.
 *
 *      st            Set to XAUTHN_S_COMPLETE, or to an error status indicating
 *                    the nature of the failure.
 */
xauthn_status_t 
xauthn_authenticate(
    xnvlist_t         *authnInfo,
    xauthn_identity_t *ident
)
{
    int               i,authmethod;
    int               prin_type;    
    int               vlen=0;
    xauthn_status_t   st;
    xattr_list_t      *xattrinfo=NULL;
    char              *authmethod_data=NULL;
    char 		* xattr_name=NULL;
    char 		* xattr_value=NULL;


    if ( authnInfo != NULL ) {
        printf("===============================\n");
        printf("User Authentication Information:\n");
        printf("===============================\n");

        for(i=0; i<authnInfo->length; i++) {
            if( ((authnInfo->items) + i)->name != NULL) {
               	printf("%s:\t%s (%d)\n", authnInfo->items[i].name,
                   		         authnInfo->items[i].value,
                                         authnInfo->items[i].vlen);
            }
        }
    }
    else {
        printf("There is NO user authenication information available.\n\n");
        return XAUTHN_S_BAD_AUTHENTICATION_INFO;
    }

   /*Determine the authorization method*/
   authmethod=UNSUPPORTED_AUTH;
   st=1; 
   if (does_name_exist_in_list(XAUTHN_EXTENDED_PARAMETER,authnInfo))
   {
      st = xnvlist_get(authnInfo,XAUTHN_EXTENDED_PARAMETER,&authmethod_data,&vlen);
   }
   if (st == 0)
     {
        if (SELECTED(authmethod_data,"basic_auth"))
	   {
           authmethod=BASIC_AUTH;
         }

       if (SELECTED(authmethod_data,"cert"))
	   {
            authmethod=CERT;
         }
     }

  /* Allow single sign-on requests to flow through */
  /* Code may be modified to check IP address or headers below */
  if (!does_name_exist_in_list(XAUTHN_USERNAME,authnInfo) &&
      !does_name_exist_in_list(XAUTHN_CERT_DN,authnInfo))
      authmethod = SSO;

   switch(authmethod)
   {
      case BASIC_AUTH:
                  st = check_basic_auth(authnInfo);
                  break;

      case CERT:
                  st = check_certdn(authnInfo);
                  break;

      case SSO:
                  st = XAUTHN_S_COMPLETE;
                  break;

      case UNSUPPORTED_AUTH:
                  if (authmethod_data)
                      free(authmethod_data);
                  return XAUTHN_S_UNSUPPORTED_AUTHN_METHOD;
   }


   /*If auth didn't succeed then return the error code*/
   if (st != XAUTHN_S_COMPLETE)
   {
    if (authmethod_data)
        free(authmethod_data);
      return st;
   }

   
  /*authentication was successful so lets return the valid AM identity*/
  /*that was passed in at init time.*/

        ident->prin.prin_type=1;
        ident->prin.data.dn=strdup(Global_demo_info.demo_validpddn);


  /*Send back 1 extended attribute*/
	xattr_name=strdup(XAUTHN_TAG_VALUE_KEY_PREFIX "Account-Number");
	xattr_value=strdup("123-456-789");

    xattrinfo = xattr_malloc(1);
    xattr_set(xattrinfo, xattr_name, xattr_value);

    ident->xattrs=*xattrinfo;

    /* Free allocated memory */
    if (xattrinfo)
        free(xattrinfo);
    if (authmethod_data)
        free(authmethod_data);

    return XAUTHN_S_COMPLETE;
}


xauthn_status_t 
xauthn_change_password(
    xnvlist_t         *authnInfo
)
{
    int               i,vlen=0;
    xauthn_status_t   st = XAUTHN_S_COMPLETE;
    char * newpassword = NULL;

    printf("====================\n");
    printf("User Password Change:\n");
    printf("====================\n");

    for(i=0; i<authnInfo->length; i++) {
        if( ((authnInfo->items) + i)->name != NULL) {
            printf("%s:\t%s (%d)\n", authnInfo->items[i].name,
                                     authnInfo->items[i].value,
                                     authnInfo->items[i].vlen);
        }
    }

  /*Search for the name value pair new password*/
   st=1; 
   if (does_name_exist_in_list(XAUTHN_NEW_PASSWORD,authnInfo))
   {
      st = xnvlist_get(authnInfo,XAUTHN_NEW_PASSWORD,&newpassword,&vlen);
   }
   if (st == 0)
    {
       /*Check to see if the password is the same as previous*/
       if (SELECTED(newpassword,Global_demo_info.demo_user_password))
        {
           st = XAUTHN_S_PASSWD_TOO_MANY_REPEATED;
        }else{
             /*Set the new password*/
             memset(Global_demo_info.demo_user_password,0,BIGDATASIZE);
             strcpy(Global_demo_info.demo_user_password,newpassword);
             st = XAUTHN_S_COMPLETE;
        }

    }else{
        st = XAUTHN_S_PASSWD_POLICY_VIOLATION;
    }

    if (newpassword)
        free(newpassword);

    return st;
}


/*
 * FUNCTION NAME
 * 	init_arg_type
 *
 * DESCRIPTION
 * 	Used ONLY for demo purposes to help parse the init args
 *
 * ARGUMENTS
 * 	[in] argtotest Argument to test.
 *
 * RETURN VALUE
 *     DEMOOTHERDATA    
 *     DEMOUSER         
 *     DEMOUSERPASS     
 *     DEMOCERTDN       
 *     VALIDPDDN
 */
int init_arg_type(char * argtotest)
{
  int type=DEMOOTHERDATA;

        if (SELECTED(argtotest,"demouser"))
		{
                type=DEMOUSER;
            }

        if (SELECTED(argtotest,"demopassword"))
            {
                type=DEMOUSERPASS;
            }

        if (SELECTED(argtotest,"democertdn"))
            {
                type=DEMOCERTDN;
            }

        if (SELECTED(argtotest,"validpddn"))
            {
                type=VALIDPDDN;
            }



   return type;
}

/*
 * FUNCTION NAME
 * 	getOptionValue()
 *
 * DESCRIPTION
 *    This function should only be called for options that are
 *    expecting a value on the command line.
 *    
 *    This function ensures that another argument exists
 *    before trying to retrieve it from the argv array. 
 *
 *    If the value exists (and is not another option)...then this routine will copy the value 
 *    from the argv array into a buffer that is used to store the value.
 *    It will also increment the option counter.
 *
 * ARGUMENTS
 * 	[in] argc     The count of arguments to the service.
 * 	[in] argv     The array of argument strings.
 *
 * RETURN VALUE
 * 	0 success
 *    1 error
 */
int getOptionValue(int * arg_num,int argc, char ** argv,char * option_buffer)
{
      int status=0;
	char * temp_ptr=NULL;
      

	if ( (*arg_num) < argc && ((*arg_num)+1) < argc) {
		if (init_arg_type(argv[(*arg_num)+1]) == DEMOOTHERDATA)
		{
              status=0;
		} else {
              status=1;
		}
	 }
	else
	{
		status=1;
	}

      /*In a production CDAS module make sure that global variables*/
      /*are protected by a locking mechanism*/
	if (status ==0)
	{
	   temp_ptr=argv[++(*arg_num)];
	}else
	{
        temp_ptr=(char *)"";

	}

	    strcpy(option_buffer,temp_ptr);

    return status;
}


/*
  FUNCTION NAME
  	check_basic_auth()
 
  DESCRIPTION
     Checks to see if User name and password are okay.

  RETURNS
     XAUTHN_S_COMPLETE                if everything is okay
     XAUTHN_S_AUTHENTICATION_FAILURE  for bad auth
*/
int check_basic_auth(xnvlist_t *authnInfo)
{
   int i,validuser=0,validpwd=0,vlen=0;
   xauthn_status_t   st=XAUTHN_S_COMPLETE;
   char * passedinusername=NULL;
   char * passedinpassword=NULL;

   st=1; 
   if (does_name_exist_in_list(XAUTHN_USERNAME,authnInfo))
   {
      st = xnvlist_get(authnInfo,XAUTHN_USERNAME,&passedinusername,&vlen);
   }
   if (st == 0)
   {
        /*compare passed in value with username passed in at init time*/
        /*Note: This is a CASE INSENSITIVE compare for demo purposes*/
        if(SELECTED(passedinusername,Global_demo_info.demo_user)) {validuser=1;}
   }

   st=1; 
   if (does_name_exist_in_list(XAUTHN_PASSWORD,authnInfo))
   {
      st = xnvlist_get(authnInfo,XAUTHN_PASSWORD,&passedinpassword,&vlen);
   }
   if (st == 0)
   {
        /*compare passed in value with username passed in at init time*/
        /*Note: This is a CASE INSENSITIVE compare for demo purposes*/
        if(SELECTED(passedinpassword,Global_demo_info.demo_user_password)){validpwd=1;}
   }

     if ( validuser==0 ||  validpwd==0 )
     {
        st = XAUTHN_S_AUTHENTICATION_FAILURE;
     }

    if (passedinusername)
        free(passedinusername);
    if (passedinpassword)
        free(passedinpassword);

    return st;
}


/*
  FUNCTION NAME
  	check_certdn()
 
  DESCRIPTION
     Checks to see if cert dn passed in is okay.

  RETURNS
     XAUTHN_S_COMPLETE                if everything is okay
     XAUTHN_S_AUTHENTICATION_FAILURE  for bad auth
*/

int check_certdn(xnvlist_t *authnInfo)
{
   int i,vlen=0,validcert=0;
   char * certdnpassedin = NULL;
   xauthn_status_t   st=XAUTHN_S_COMPLETE;

   st=1; 
   if (does_name_exist_in_list(XAUTHN_CERT_DN,authnInfo))
   {
      st = xnvlist_get(authnInfo,XAUTHN_CERT_DN,&certdnpassedin,&vlen);
   }
   if (st == 0)
   {
        /*compare passed in value with cert dn passed in at init time*/
        /*Note: This is a CASE INSENSITIVE compare for demo purposes*/
        if(SELECTED(certdnpassedin,Global_demo_info.demo_cert_dn)) {validcert=1;}
    }

     if ( validcert==0)
     {
        st = XAUTHN_S_CERTIFICATE_AUTH_FAILURE;
     }

    if (certdnpassedin)
        free(certdnpassedin);

    return st;
}


/*
  FUNCTION NAME
  	does_name_exist_in_list()
 
  DESCRIPTION
     Checks to see if a particular name exists in the list.
     If xnvlist_get is called on a name that doesn't exist then xnvlist_get will
     cause a segmentation fault.
  RETURNS
     0  Does Not Exist
     1  Does Exist
*/

int does_name_exist_in_list(char * name_to_check,xnvlist_t *authnInfo)
{
   int i;
   xauthn_status_t   st=0;

    for(i=0; i<authnInfo->length; i++) {
        if( ((authnInfo->items) + i)->name != NULL) {
              if(SELECTED(((authnInfo->items) + i)->name,name_to_check)) {st=1;}
        }
      if (st ==1 ) {break;}
    }

    return st;
}